from typing import Any, List, Optional
from deprecated import deprecated

from specklepy.objects.geometry import Point, Vector

from .base import Base

OTHER = "Objects.Other."
OTHER_REVIT = OTHER + "Revit."

IDENTITY_TRANSFORM = [
    1.0,
    0.0,
    0.0,
    0.0,
    0.0,
    1.0,
    0.0,
    0.0,
    0.0,
    0.0,
    1.0,
    0.0,
    0.0,
    0.0,
    0.0,
    1.0,
]


class Material(Base, speckle_type=OTHER + "Material"):
    """Generic class for materials containing generic parameters."""

    name: Optional[str] = None


class RevitMaterial(Material, speckle_type="Objects.Other.Revit." + "RevitMaterial"):
    materialCategory: Optional[str] = None
    materialClass: Optional[str] = None
    shininess: Optional[int] = None
    smoothness: Optional[int] = None
    transparency: Optional[int] = None
    parameters: Optional[Base] = None


class RenderMaterial(Base, speckle_type=OTHER + "RenderMaterial"):
    name: Optional[str] = None
    opacity: float = 1
    metalness: float = 0
    roughness: float = 1
    diffuse: int = -2894893  # light gray arbg
    emissive: int = -16777216  # black arbg


class MaterialQuantity(Base, speckle_type=OTHER + "MaterialQuantity"):
    material: Optional[Material] = None
    volume: Optional[float] = None
    area: Optional[float] = None


class DisplayStyle(Base, speckle_type=OTHER + "DisplayStyle"):
    """
    Minimal display style class.
    Developed primarily for display styles in Rhino and AutoCAD.
    Rhino object attributes uses OpenNURBS definition for linetypes and lineweights.
    """

    name: Optional[str] = None
    color: int = -2894893  # light gray arbg
    linetype: Optional[str] = None
    lineweight: float = 0


class Transform(
    Base,
    speckle_type=OTHER + "Transform",
    serialize_ignore={"translation", "scaling", "is_identity", "value"},
):
    """The 4x4 transformation matrix

    The 3x3 sub-matrix determines scaling.
    The 4th column defines translation,
    where the last value is a divisor (usually equal to 1).
    """

    _value: Optional[List[float]] = None

    @property
    @deprecated(version="2.12", reason="Use matrix")
    def value(self) -> List[float]:
        return self._value

    @value.setter
    def value(self, value: List[float]) -> None:
        self.matrix = value

    @property
    def matrix(self) -> List[float]:
        """The transform matrix represented as a flat list of 16 floats"""
        return self._value

    @matrix.setter
    def matrix(self, value: List[float]) -> None:
        try:
            value = [float(x) for x in value]
        except (ValueError, TypeError) as error:
            raise ValueError(
                "Could not create a Transform object with the requested value. Input"
                f" must be a 16 element list of numbers. Value provided: {value}"
            ) from error

        if len(value) != 16:
            raise ValueError(
                "Could not create a Transform object: input list should be 16 floats"
                f" long, but was {len(value)} long"
            )

        self._value = value

    @property
    def translation(self) -> List[float]:
        """The final column of the matrix which defines the translation"""
        return [self._value[i] for i in (3, 7, 11, 15)]

    @property
    def scaling(self) -> List[float]:
        """The 3x3 scaling sub-matrix"""
        return [self._value[i] for i in (0, 1, 2, 4, 5, 6, 8, 9, 10)]

    @property
    def is_identity(self) -> bool:
        return self._value == IDENTITY_TRANSFORM

    def apply_to_point(self, point: Point) -> Point:
        """Transform a single speckle Point

        Arguments:
            point {Point} -- the speckle Point to transform

        Returns:
            Point -- a new transformed point
        """
        coords = self.apply_to_point_value([point.x, point.y, point.z])
        return Point(x=coords[0], y=coords[1], z=coords[2], units=point.units)

    def apply_to_point_value(self, point_value: List[float]) -> List[float]:
        """Transform a list of three floats representing a point

        Arguments:
            point_value {List[float]} -- a list of 3 floats

        Returns:
            List[float] -- the list with the transform applied
        """
        transformed = [
            point_value[0] * self._value[i]
            + point_value[1] * self._value[i + 1]
            + point_value[2] * self._value[i + 2]
            + self._value[i + 3]
            for i in range(0, 15, 4)
        ]

        return [transformed[i] / transformed[3] for i in range(3)]

    def apply_to_points(self, points: List[Point]) -> List[Point]:
        """Transform a list of speckle Points

        Arguments:
            points {List[Point]} -- the list of speckle Points to transform

        Returns:
            List[Point] -- a new list of transformed points
        """
        return [self.apply_to_point(point) for point in points]

    def apply_to_points_values(self, points_value: List[float]) -> List[float]:
        """Transform a list of speckle Points

        Arguments:
            points {List[float]}
            -- a flat list of floats representing points to transform

        Returns:
            List[float] -- a new transformed list
        """
        if len(points_value) % 3 != 0:
            raise ValueError(
                "Cannot apply transform as the points list is malformed: expected"
                " length to be multiple of 3"
            )
        transformed = []
        for i in range(0, len(points_value), 3):
            transformed.extend(self.apply_to_point_value(points_value[i : i + 3]))

        return transformed

    def apply_to_vector(self, vector: Vector) -> Vector:
        """Transform a single speckle Vector

        Arguments:
            point {Vector} -- the speckle Vector to transform

        Returns:
            Vector -- a new transformed point
        """
        coords = self.apply_to_vector_value([vector.x, vector.y, vector.z])
        return Vector(x=coords[0], y=coords[1], z=coords[2], units=vector.units)

    def apply_to_vector_value(self, vector_value: List[float]) -> List[float]:
        """Transform a list of three floats representing a vector

        Arguments:
            vector_value {List[float]} -- a list of 3 floats

        Returns:
            List[float] -- the list with the transform applied
        """
        return [
            vector_value[0] * self._value[i]
            + vector_value[1] * self._value[i + 1]
            + vector_value[2] * self._value[i + 2]
            for i in range(0, 15, 4)
        ][:3]

    @classmethod
    def from_list(cls, value: Optional[List[float]] = None) -> "Transform":
        """Returns a Transform object from a list of 16 numbers.
        If no value is provided, an identity transform will be returned.

        Arguments:
            value {List[float]} -- the matrix as a flat list of 16 numbers
            (defaults to the identity transform)

        Returns:
            Transform -- a complete transform object
        """
        if not value:
            value = IDENTITY_TRANSFORM
        return cls(value=value)


class BlockDefinition(
    Base, speckle_type=OTHER + "BlockDefinition", detachable={"geometry"}
):
    name: Optional[str] = None
    basePoint: Optional[Point] = None
    geometry: Optional[List[Base]] = None


class Instance(
    Base, speckle_type=OTHER + "Instance", detachable={"definition"}
):
    transform: Optional[Transform] = None
    definition: Optional[Base] = None


class BlockInstance(
    Instance, speckle_type=OTHER + "BlockInstance", serialize_ignore={"blockDefinition"}
):
    @property
    @deprecated(version="2.13", reason="Use definition")
    def blockDefinition(self) -> Optional[BlockDefinition]:
        if isinstance(self.definition, BlockDefinition):
            return self.definition
        return None

    @blockDefinition.setter
    def blockDefinition(self, value: Optional[BlockDefinition]) -> None:
        self.definition = value

class RevitInstance(Instance, speckle_type=OTHER_REVIT + "RevitInstance"):
    level: Optional[Base]  = None
    facingFlipped: bool
    handFlipped: bool
    parameters: Optional[Base]  = None
    elementId: Optional[str]

# TODO: prob move this into a built elements module, but just trialling this for now
class RevitParameter(
    Base, speckle_type="Objects.BuiltElements.Revit.Parameter"
):
    name: Optional[str] = None
    value: Any = None
    applicationUnitType: Optional[str] = None  # eg UnitType UT_Length
    applicationUnit: Optional[str] = None  # DisplayUnitType eg DUT_MILLIMITERS
    applicationInternalName: Optional[
        str
    ] = None  # BuiltInParameterName or GUID for shared parameter
    isShared: bool = False
    isReadOnly: bool = False
    isTypeParameter: bool = False

class Collection(
    Base, speckle_type="Speckle.Core.Models.Collection", detachable={"elements"}
):
    name: Optional[str] = None
    collectionType: Optional[str] = None
    elements: Optional[List[Base]] = None


from datetime import datetime, timezone
from typing import List, Optional, Union

from gql import gql

from specklepy.api.models import ActivityCollection, LimitedUser
from specklepy.api.resource import ResourceBase
from specklepy.logging import metrics
from specklepy.logging.exceptions import SpeckleException

from specklepy.core.api.resources.other_user import Resource as CoreResource


class Resource(CoreResource):
    """API Access class for other users, that are not the currently active user."""

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            server_version=server_version,
        )
        self.schema = LimitedUser

    def get(self, id: str) -> LimitedUser:
        """
        Gets the profile of another user.

        Arguments:
            id {str} -- the user id

        Returns:
            LimitedUser -- the retrieved profile of another user
        """
        metrics.track(metrics.SDK, self.account, {"name": "Other User Get"})
        return super().get(id)

    def search(
        self, search_query: str, limit: int = 25
    ) -> Union[List[LimitedUser], SpeckleException]:
        """Searches for user by name or email. The search query must be at least
        3 characters long

        Arguments:
            search_query {str} -- a string to search for
            limit {int} -- the maximum number of results to return
        Returns:
            List[LimitedUser] -- a list of User objects that match the search query
        """
        if len(search_query) < 3:
            return SpeckleException(
                message="User search query must be at least 3 characters"
            )

        metrics.track(metrics.SDK, self.account, {"name": "Other User Search"})
        return super().search(search_query, limit) 

    def activity(
        self,
        user_id: str,
        limit: int = 20,
        action_type: Optional[str] = None,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
        cursor: Optional[datetime] = None,
    ) -> ActivityCollection:
        """
        Get the activity from a given stream in an Activity collection.
        Step into the activity `items` for the list of activity.

        Note: all timestamps arguments should be `datetime` of
        any tz as they will be converted to UTC ISO format strings

        user_id {str} -- the id of the user to get the activity from
        action_type {str} -- filter results to a single action type
            (eg: `commit_create` or `commit_receive`)
        limit {int} -- max number of Activity items to return
        before {datetime} -- latest cutoff for activity
            (ie: return all activity _before_ this time)
        after {datetime} -- oldest cutoff for activity
            (ie: return all activity _after_ this time)
        cursor {datetime} -- timestamp cursor for pagination
        """
        metrics.track(metrics.SDK, self.account, {"name": "Other User Activity"})
        return super().activity(user_id, limit, action_type, before, after, cursor) 
    


from threading import Lock
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from gql.client import Client
from gql.transport.exceptions import TransportQueryError
from graphql import DocumentNode

from specklepy.api.credentials import Account
from specklepy.logging.exceptions import (
    GraphQLException,
    SpeckleException,
    UnsupportedException,
)
from specklepy.serialization.base_object_serializer import BaseObjectSerializer
from specklepy.transports.sqlite import SQLiteTransport

# following imports seem to be unnecessary, but they need to stay 
# to not break the scripts using these functions as non-core
from specklepy.core.api.resource import ResourceBase as CoreResourceBase


class ResourceBase(CoreResourceBase):
    def __init__(
        self,
        account: Account,
        basepath: str,
        client: Client,
        name: str,
        server_version: Optional[Tuple[Any, ...]] = None,
    ) -> None:
        super().__init__(
            account = account,
            basepath = basepath,
            client = client,
            name = name,
            server_version = server_version
        )
    


from typing import Optional

from gql import gql

from specklepy.core.api.models import Branch
from specklepy.core.api.resource import ResourceBase

NAME = "branch"


class Resource(ResourceBase):
    """API Access class for branches"""

    def __init__(self, account, basepath, client) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            name=NAME,
        )
        self.schema = Branch

    def create(
        self, stream_id: str, name: str, description: str = "No description provided"
    ) -> str:
        """Create a new branch on this stream

        Arguments:
            name {str} -- the name of the new branch
            description {str} -- a short description of the branch

        Returns:
            id {str} -- the newly created branch's id
        """
        query = gql(
            """
            mutation BranchCreate($branch: BranchCreateInput!) {
              branchCreate(branch: $branch)
            }
          """
        )
        params = {
            "branch": {
                "streamId": stream_id,
                "name": name,
                "description": description,
            }
        }

        return self.make_request(
            query=query, params=params, return_type="branchCreate", parse_response=False
        )

    def get(self, stream_id: str, name: str, commits_limit: int = 10):
        """Get a branch by name from a stream

        Arguments:
            stream_id {str} -- the id of the stream to get the branch from
            name {str} -- the name of the branch to get
            commits_limit {int} -- maximum number of commits to get

        Returns:
            Branch -- the fetched branch with its latest commits
        """
        query = gql(
            """
            query BranchGet($stream_id: String!, $name: String!, $commits_limit: Int!) {
                stream(id: $stream_id) {
                        branch(name: $name) {
                          id,
                          name,
                          description,
                          commits (limit: $commits_limit) {
                            totalCount,
                            cursor,
                            items {
                              id,
                              referencedObject,
                              sourceApplication,
                              totalChildrenCount,
                              message,
                              authorName,
                              authorId,
                              branchName,
                              parents,
                              createdAt
                            }
                        }
                    }
                }
            }
            """
        )

        params = {"stream_id": stream_id, "name": name, "commits_limit": commits_limit}

        return self.make_request(
            query=query, params=params, return_type=["stream", "branch"]
        )

    def list(self, stream_id: str, branches_limit: int = 10, commits_limit: int = 10):
        """Get a list of branches from a given stream

        Arguments:
            stream_id {str} -- the id of the stream to get the branches from
            branches_limit {int} -- maximum number of branches to get
            commits_limit {int} -- maximum number of commits to get

        Returns:
            List[Branch] -- the branches on the stream
        """
        query = gql(
            """
            query BranchesGet(
                    $stream_id: String!,
                    $branches_limit: Int!,
                    $commits_limit: Int!
                ) {
                stream(id: $stream_id) {
                    branches(limit: $branches_limit) {
                        items {
                            id
                            name
                            description
                            commits(limit: $commits_limit) {
                                totalCount
                                items{
                                    id
                                    message
                                    referencedObject
                                    sourceApplication
                                    parents
                                    authorId
                                    authorName
                                    branchName
                                    createdAt
                                }
                            }
                        }
                    }
                }
            }
            """
        )

        params = {
            "stream_id": stream_id,
            "branches_limit": branches_limit,
            "commits_limit": commits_limit,
        }

        return self.make_request(
            query=query, params=params, return_type=["stream", "branches", "items"]
        )

    def update(
        self,
        stream_id: str,
        branch_id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ):
        """Update a branch

        Arguments:
            stream_id {str} -- the id of the stream containing the branch to update
            branch_id {str} -- the id of the branch to update
            name {str} -- optional: the updated branch name
            description {str} -- optional: the updated branch description

        Returns:
            bool -- True if update is successful
        """
        query = gql(
            """
            mutation  BranchUpdate($branch: BranchUpdateInput!) {
                branchUpdate(branch: $branch)
                }
            """
        )
        params = {
            "branch": {
                "streamId": stream_id,
                "id": branch_id,
            }
        }

        if name:
            params["branch"]["name"] = name
        if description:
            params["branch"]["description"] = description

        return self.make_request(
            query=query, params=params, return_type="branchUpdate", parse_response=False
        )

    def delete(self, stream_id: str, branch_id: str):
        """Delete a branch

        Arguments:
            stream_id {str} -- the id of the stream containing the branch to delete
            branch_id {str} -- the branch to delete

        Returns:
            bool -- True if deletion is successful
        """
        query = gql(
            """
            mutation BranchDelete($branch: BranchDeleteInput!) {
                branchDelete(branch: $branch)
            }
            """
        )

        params = {"branch": {"streamId": stream_id, "id": branch_id}}

        return self.make_request(
            query=query, params=params, return_type="branchDelete", parse_response=False
        )



"""Function execution module.

Provides mechanisms to execute any function,
 that conforms to the AutomateFunction "interface"
"""
import json
import os
import sys
import traceback
from pathlib import Path
from typing import Callable, Optional, TypeVar, Union, overload

from speckle_automate.automation_context import AutomationContext
from speckle_automate.schema import AutomateBase, AutomationRunData, AutomationStatus

T = TypeVar("T", bound=AutomateBase)

AutomateFunction = Callable[[AutomationContext, T], None]
AutomateFunctionWithoutInputs = Callable[[AutomationContext], None]


@overload
def execute_automate_function(
    automate_function: AutomateFunction[T],
    input_schema: type[T],
) -> None:
    ...


@overload
def execute_automate_function(automate_function: AutomateFunctionWithoutInputs) -> None:
    ...


def execute_automate_function(
    automate_function: Union[AutomateFunction[T], AutomateFunctionWithoutInputs],
    input_schema: Optional[type[T]] = None,
):
    """Runs the provided automate function with the input schema."""
    # first arg is the python file name, we do not need that
    args = sys.argv[1:]

    if len(args) < 2:
        raise ValueError("too few arguments specified need minimum 2")

    if len(args) > 4:
        raise ValueError("too many arguments specified, max supported is 4")

    # we rely on a command name convention to decide what to do.
    # this is here, so that the function authors do not see any of this
    command = args[0]

    if command == "generate_schema":
        path = Path(args[1])
        schema = json.dumps(
            input_schema.model_json_schema(by_alias=True) if input_schema else {}
        )
        path.write_text(schema)

    elif command == "run":
        automation_run_data = args[1]
        function_inputs = args[2]

        speckle_token = os.environ.get("SPECKLE_TOKEN", None)
        if not speckle_token and len(args) != 4:
            raise ValueError("Cannot get speckle token from arguments or environment")

        speckle_token = speckle_token if speckle_token else args[3]
        automation_context = AutomationContext.initialize(
            automation_run_data, speckle_token
        )

        inputs = (
            input_schema.model_validate_json(function_inputs)
            if input_schema
            else input_schema
        )

        if inputs:
            automation_context = run_function(
                automation_context,
                automate_function,  # type: ignore
                inputs,
            )
        else:
            automation_context = run_function(
                automation_context,
                automate_function,  # type: ignore
            )

        exit_code = (
            0 if automation_context.run_status == AutomationStatus.SUCCEEDED else 1
        )
        exit(exit_code)

    else:
        raise NotImplementedError(f"Command: '{command}' is not supported.")


@overload
def run_function(
    automation_context: AutomationContext,
    automate_function: AutomateFunction[T],
    inputs: T,
) -> AutomationContext:
    ...


@overload
def run_function(
    automation_context: AutomationContext,
    automate_function: AutomateFunctionWithoutInputs,
) -> AutomationContext:
    ...


def run_function(
    automation_context: AutomationContext,
    automate_function: Union[AutomateFunction[T], AutomateFunctionWithoutInputs],
    inputs: Optional[T] = None,
) -> AutomationContext:
    """Run the provided function with the automate sdk context."""
    automation_context.report_run_status()

    try:
        # avoiding complex type gymnastics here on the internals.
        # the external type overloads make this correct
        if inputs:
            automate_function(automation_context, inputs)  # type: ignore
        else:
            automate_function(automation_context)  # type: ignore

        # the function author forgot to mark the function success
        if automation_context.run_status not in [
            AutomationStatus.FAILED,
            AutomationStatus.SUCCEEDED,
        ]:
            automation_context.mark_run_success(
                "WARNING: Automate assumed a success status,"
                " but it was not marked as so by the function."
            )
    except Exception:
        trace = traceback.format_exc()
        print(trace)
        automation_context.mark_run_failed(
            "Function error. Check the automation run logs for details."
        )
    finally:
        automation_context.report_run_status()
        return automation_context



"""Builtin Speckle object kit."""

from specklepy.objects.structural.analysis import (
    Model,
    ModelInfo,
    ModelSettings,
    ModelUnits,
)
from specklepy.objects.structural.axis import (
    AxisType,
    Axis
)
from specklepy.objects.structural.geometry import (
    Element1D,
    Element2D,
    Element3D,
    ElementType1D,
    ElementType2D,
    ElementType3D,
    Node,
    Restraint,
)
from specklepy.objects.structural.loading import (
    ActionType,
    BeamLoadType,
    CombinationType,
    FaceLoadType,
    Load,
    LoadAxisType,
    LoadBeam,
    LoadCase,
    LoadCombinations,
    LoadDirection,
    LoadDirection2D,
    LoadFace,
    LoadGravity,
    LoadNode,
    LoadType,
)
from specklepy.objects.structural.materials import (
    Concrete,
    MaterialType,
    Steel,
    StructuralMaterial,
    Timber,
)
from specklepy.objects.structural.properties import (
    BaseReferencePoint,
    MemberType,
    Property,
    Property1D,
    Property2D,
    Property3D,
    PropertyDamper,
    PropertyMass,
    PropertySpring,
    PropertyType2D,
    PropertyType3D,
    PropertyTypeDamper,
    PropertyTypeSpring,
    ReferenceSurface,
    ReferenceSurfaceEnum,
    SectionProfile,
    ShapeType,
    shapeType,
)
from specklepy.objects.structural.results import (
    Result,
    Result1D,
    Result2D,
    Result3D,
    ResultGlobal,
    ResultNode,
    ResultSet1D,
    ResultSet2D,
    ResultSet3D,
    ResultSetAll,
    ResultSetNode,
)

__all__ = [
    "Element1D",
    "Element2D",
    "Element3D",
    "ElementType1D",
    "ElementType2D",
    "ElementType3D",
    "AxisType",
    "Axis",
    "Node",
    "Restraint",
    "Load",
    "LoadType",
    "ActionType",
    "BeamLoadType",
    "FaceLoadType",
    "LoadDirection",
    "LoadDirection2D",
    "LoadAxisType",
    "CombinationType",
    "LoadBeam",
    "LoadCase",
    "LoadCombinations",
    "LoadFace",
    "LoadGravity",
    "LoadNode",
    "Model",
    "ModelInfo",
    "ModelSettings",
    "ModelUnits",
    "MaterialType",
    "Concrete",
    "StructuralMaterial",
    "Steel",
    "Timber",
    "Property",
    "Property1D",
    "Property2D",
    "Property3D",
    "PropertyDamper",
    "PropertyMass",
    "PropertySpring",
    "SectionProfile",
    "MemberType",
    "BaseReferencePoint",
    "ReferenceSurface",
    "PropertyType2D",
    "PropertyType3D",
    "ShapeType",
    "PropertyTypeSpring",
    "PropertyTypeDamper",
    "ReferenceSurfaceEnum",
    "shapeType",
    "Result",
    "Result1D",
    "ResultSet1D",
    "Result2D",
    "ResultSet2D",
    "Result3D",
    "ResultSet3D",
    "ResultGlobal",
    "ResultSetNode",
    "ResultNode",
    "ResultSetAll",
]



"""Some useful helpers for working with automation data."""
import secrets
import string

from specklepy.api.client import SpeckleClient
from gql import gql


def register_new_automation(
    speckle_client: SpeckleClient,
    project_id: str,
    model_id: str,
    automation_id: str,
    automation_name: str,
    automation_revision_id: str,
) -> bool:
    """Register a new automation in the speckle server."""
    query = gql(
        """
        mutation CreateAutomation(
            $projectId: String! 
            $modelId: String! 
            $automationName: String!
            $automationId: String! 
            $automationRevisionId: String!
        ) {
                automationMutations {
                    create(
                        input: {
                            projectId: $projectId
                            modelId: $modelId
                            automationName: $automationName 
                            automationId: $automationId
                            automationRevisionId: $automationRevisionId
                        }
                    )
                }
            }
        """
    )
    params = {
        "projectId": project_id,
        "modelId": model_id,
        "automationName": automation_name,
        "automationId": automation_id,
        "automationRevisionId": automation_revision_id,
    }
    return speckle_client.httpclient.execute(query, params)


def crypto_random_string(length: int) -> str:
    """Generate a semi crypto random string of a given length."""
    alphabet = string.ascii_letters + string.digits
    return "".join(secrets.choice(alphabet) for _ in range(length)).lower()



import re
from typing import Any, Dict, List, Tuple

from gql import gql

from specklepy.api.models import ServerInfo
from specklepy.api.resource import ResourceBase
from specklepy.logging import metrics
from specklepy.logging.exceptions import GraphQLException

from specklepy.core.api.resources.server import Resource as CoreResource


class Resource(CoreResource):
    """API Access class for the server"""

    def __init__(self, account, basepath, client) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
        )

    def get(self) -> ServerInfo:
        """Get the server info

        Returns:
            dict -- the server info in dictionary form
        """
        metrics.track(metrics.SDK, self.account, {"name": "Server Get"})
        return super().get()

    def version(self) -> Tuple[Any, ...]:
        """Get the server version

        Returns:
            the server version in the format (major, minor, patch, (tag, build))
            eg (2, 6, 3) for a stable build and (2, 6, 4, 'alpha', 4711) for alpha
        """
        # not tracking as it will be called along with other mutations / queries as a check
        return super().version()

    def apps(self) -> Dict:
        """Get the apps registered on the server

        Returns:
            dict -- a dictionary of apps registered on the server
        """
        metrics.track(metrics.SDK, self.account, {"name": "Server Apps"})
        return super().apps()

    def create_token(self, name: str, scopes: List[str], lifespan: int) -> str:
        """Create a personal API token

        Arguments:
            scopes {List[str]} -- the scopes to grant with this token
            name {str} -- a name for your new token
            lifespan {int} -- duration before the token expires

        Returns:
            str -- the new API token. note: this is the only time you'll see the token!
        """
        metrics.track(metrics.SDK, self.account, {"name": "Server Create Token"})
        return super().create_token(name, scopes, lifespan)

    def revoke_token(self, token: str) -> bool:
        """Revokes (deletes) a personal API token

        Arguments:
            token {str} -- the token to revoke (delete)

        Returns:
            bool -- True if the token was successfully deleted
        """
        metrics.track(metrics.SDK, self.account, {"name": "Server Revoke Token"})
        return super().revoke_token(token) 



from datetime import datetime, timezone
from typing import List, Optional

from deprecated import deprecated
from gql import gql

from specklepy.api.models import ActivityCollection, PendingStreamCollaborator, Stream
from specklepy.api.resource import ResourceBase
from specklepy.logging import metrics
from specklepy.logging.exceptions import SpeckleException, UnsupportedException

from specklepy.core.api.resources.stream import Resource as CoreResource


class Resource(CoreResource):
    """API Access class for streams"""

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            server_version=server_version,
        )

        self.schema = Stream

    def get(self, id: str, branch_limit: int = 10, commit_limit: int = 10) -> Stream:
        """Get the specified stream from the server

        Arguments:
            id {str} -- the stream id
            branch_limit {int} -- the maximum number of branches to return
            commit_limit {int} -- the maximum number of commits to return

        Returns:
            Stream -- the retrieved stream
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Get"})
        return super().get(id, branch_limit, commit_limit)

    def list(self, stream_limit: int = 10) -> List[Stream]:
        """Get a list of the user's streams

        Arguments:
            stream_limit {int} -- The maximum number of streams to return

        Returns:
            List[Stream] -- A list of Stream objects
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream List"})
        return super().list(stream_limit)

    def create(
        self,
        name: str = "Anonymous Python Stream",
        description: str = "No description provided",
        is_public: bool = True,
    ) -> str:
        """Create a new stream

        Arguments:
            name {str} -- the name of the string
            description {str} -- a short description of the stream
            is_public {bool}
                -- whether or not the stream can be viewed by anyone with the id

        Returns:
            id {str} -- the id of the newly created stream
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Create"})
        return super().create(name, description, is_public)

    def update(
        self,
        id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        is_public: Optional[bool] = None,
    ) -> bool:
        """Update an existing stream

        Arguments:
            id {str} -- the id of the stream to be updated
            name {str} -- the name of the string
            description {str} -- a short description of the stream
            is_public {bool}
                -- whether or not the stream can be viewed by anyone with the id

        Returns:
            bool -- whether the stream update was successful
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Update"})
        return super().update(id, name, description, is_public)

    def delete(self, id: str) -> bool:
        """Delete a stream given its id

        Arguments:
            id {str} -- the id of the stream to delete

        Returns:
            bool -- whether the deletion was successful
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Delete"})
        return super().delete(id)

    def search(
        self,
        search_query: str,
        limit: int = 25,
        branch_limit: int = 10,
        commit_limit: int = 10,
    ):
        """Search for streams by name, description, or id

        Arguments:
            search_query {str} -- a string to search for
            limit {int} -- the maximum number of results to return
            branch_limit {int} -- the maximum number of branches to return
            commit_limit {int} -- the maximum number of commits to return

        Returns:
            List[Stream] -- a list of Streams that match the search query
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Search"})
        return super().search(search_query, limit, branch_limit, commit_limit)

    def favorite(self, stream_id: str, favorited: bool = True):
        """Favorite or unfavorite the given stream.

        Arguments:
            stream_id {str} -- the id of the stream to favorite / unfavorite
            favorited {bool}
                -- whether to favorite (True) or unfavorite (False) the stream

        Returns:
            Stream -- the stream with its `id`, `name`, and `favoritedDate`
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Favorite"})
        return super().favorite(stream_id, favorited)

    def get_all_pending_invites(
        self, stream_id: str
    ) -> List[PendingStreamCollaborator]:
        """Get all of the pending invites on a stream.
        You must be a `stream:owner` to query this.

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the stream id from which to get the pending invites

        Returns:
            List[PendingStreamCollaborator]
                -- a list of pending invites for the specified stream
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Invite Get"})
        return super().get_all_pending_invites(stream_id)

    def invite(
        self,
        stream_id: str,
        email: Optional[str] = None,
        user_id: Optional[str] = None,
        role: str = "stream:contributor",  # should default be reviewer?
        message: Optional[str] = None,
    ):
        """Invite someone to a stream using either their email or user id

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the id of the stream to invite the user to
            email {str} -- the email of the user to invite (use this OR `user_id`)
            user_id {str} -- the id of the user to invite (use this OR `email`)
            role {str}
                -- the role to assign to the user (defaults to `stream:contributor`)
            message {str}
                -- a message to send along with this invite to the specified user

        Returns:
            bool -- True if the operation was successful
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Invite Create"})
        return super().invite(stream_id, email, user_id, role, message)

    def invite_batch(
        self,
        stream_id: str,
        emails: Optional[List[str]] = None,
        user_ids: Optional[List[None]] = None,
        message: Optional[str] = None,
    ) -> bool:
        """Invite a batch of users to a specified stream.

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the id of the stream to invite the user to
            emails {List[str]}
                -- the email of the user to invite (use this and/or `user_ids`)
            user_id {List[str]}
                -- the id of the user to invite (use this and/or `emails`)
            message {str}
                -- a message to send along with this invite to the specified user

        Returns:
            bool -- True if the operation was successful
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Invite Batch Create"})
        return super().invite_batch(stream_id, emails, user_ids, message)

    def invite_cancel(self, stream_id: str, invite_id: str) -> bool:
        """Cancel an existing stream invite

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the id of the stream invite
            invite_id {str} -- the id of the invite to use

        Returns:
            bool -- true if the operation was successful
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Invite Cancel"})
        return super().invite_cancel(stream_id, invite_id)

    def invite_use(self, stream_id: str, token: str, accept: bool = True) -> bool:
        """Accept or decline a stream invite

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str}
                -- the id of the stream for which the user has a pending invite
            token {str} -- the token of the invite to use
            accept {bool} -- whether or not to accept the invite (defaults to True)

        Returns:
            bool -- true if the operation was successful
        """
        metrics.track(metrics.SDK, self.account, {"name": "Invite Use"})
        return super().invite_use(stream_id, token, accept)

    def update_permission(self, stream_id: str, user_id: str, role: str):
        """Updates permissions for a user on a given stream

        Valid for Speckle Server >=2.6.4

        Arguments:
            stream_id {str} -- the id of the stream to grant permissions to
            user_id {str} -- the id of the user to grant permissions for
            role {str} -- the role to grant the user

        Returns:
            bool -- True if the operation was successful
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Permission Update", "role": role})
        return super().update_permission(stream_id, user_id, role)

    def revoke_permission(self, stream_id: str, user_id: str):
        """Revoke permissions from a user on a given stream

        Arguments:
            stream_id {str} -- the id of the stream to revoke permissions from
            user_id {str} -- the id of the user to revoke permissions from

        Returns:
            bool -- True if the operation was successful
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Permission Revoke"})
        return super().revoke_permission(stream_id, user_id)

    def activity(
        self,
        stream_id: str,
        action_type: Optional[str] = None,
        limit: int = 20,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
        cursor: Optional[datetime] = None,
    ):
        """
        Get the activity from a given stream in an Activity collection.
        Step into the activity `items` for the list of activity.

        Note: all timestamps arguments should be `datetime` of any tz
            as they will be converted to UTC ISO format strings

        stream_id {str} -- the id of the stream to get activity from
        action_type {str}
            -- filter results to a single action type
            (eg: `commit_create` or `commit_receive`)
        limit {int} -- max number of Activity items to return
        before {datetime}
            -- latest cutoff for activity (ie: return all activity _before_ this time)
        after {datetime}
            -- oldest cutoff for activity (ie: return all activity _after_ this time)
        cursor {datetime} -- timestamp cursor for pagination
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Activity"})
        return super().activity(stream_id, action_type, limit, before, after, cursor)
    


from enum import Enum
from typing import Optional

from specklepy.objects.base import Base
from specklepy.objects.structural.axis import Axis
from specklepy.objects.structural.materials import StructuralMaterial

STRUCTURAL_PROPERTY = "Objects.Structural.Properties"


class MemberType(int, Enum):
    Beam = 0
    Column = 1
    Generic1D = 2
    Slab = 3
    Wall = 4
    Generic2D = 5
    VoidCutter1D = 6
    VoidCutter2D = 7


class BaseReferencePoint(int, Enum):
    Centroid = 0
    TopLeft = 1
    TopCentre = 2
    TopRight = 3
    MidLeft = 4
    MidRight = 5
    BotLeft = 6
    BotCentre = 7
    BotRight = 8


class ReferenceSurface(int, Enum):
    Top = 0
    Middle = 1
    Bottom = 2


class PropertyType2D(int, Enum):
    Stress = 0
    Fabric = 1
    Plate = 2
    Shell = 3
    Curved = 4
    Wall = 5
    Strain = 6
    Axi = 7
    Load = 8


class PropertyType3D(int, Enum):
    Solid = 0
    Infinite = 1


class ShapeType(int, Enum):
    Rectangular = 0
    Circular = 1
    I = 2  # noqa: E741
    Tee = 3
    Angle = 4
    Channel = 5
    Perimeter = 6
    Box = 7
    Catalogue = 8
    Explicit = 9
    Undefined = 10


class PropertyTypeSpring(int, Enum):
    Axial = 0
    Torsional = 1
    General = 2
    Matrix = 3
    TensionOnly = 4
    CompressionOnly = 5
    Connector = 6
    LockUp = 7
    Gap = 8
    Friction = 9


class PropertyTypeDamper(int, Enum):
    Axial = 0
    Torsional = 1
    General = 2


class Property(Base, speckle_type=STRUCTURAL_PROPERTY):
    name: Optional[str] = None


class SectionProfile(
    Base, speckle_type=STRUCTURAL_PROPERTY + ".Profiles.SectionProfile"
):
    name: Optional[str] = None
    shapeType: Optional[ShapeType] = None
    area: float = 0.0
    Iyy: float = 0.0
    Izz: float = 0.0
    J: float = 0.0
    Ky: float = 0.0
    weight: float = 0.0


class Property1D(Property, speckle_type=STRUCTURAL_PROPERTY + ".Property1D"):
    memberType: Optional[MemberType] = None
    material: Optional[StructuralMaterial] = None
    profile: Optional[SectionProfile] = None
    referencePoint: Optional[BaseReferencePoint] = None
    offsetY: float = 0.0
    offsetZ: float = 0.0


class Property2D(Property, speckle_type=STRUCTURAL_PROPERTY + ".Property2D"):
    type: Optional[PropertyType2D] = None
    thickness: float = 0.0
    material: Optional[StructuralMaterial] = None
    orientationAxis: Optional[Axis] = None
    refSurface: Optional[ReferenceSurface] = None
    zOffset: float = 0.0
    modifierInPlane: float = 0.0
    modifierBending: float = 0.0
    modifierShear: float = 0.0
    modifierVolume: float = 0.0


class Property3D(Property, speckle_type=STRUCTURAL_PROPERTY + ".Property3D"):
    type: Optional[PropertyType3D] = None
    material: Optional[StructuralMaterial] = None
    orientationAxis: Optional[Axis] = None


class PropertyDamper(Property, speckle_type=STRUCTURAL_PROPERTY + ".PropertyDamper"):
    damperType: Optional[PropertyTypeDamper] = None
    dampingX: float = 0.0
    dampingY: float = 0.0
    dampingZ: float = 0.0
    dampingXX: float = 0.0
    dampingYY: float = 0.0
    dampingZZ: float = 0.0


class PropertyMass(Property, speckle_type=STRUCTURAL_PROPERTY + ".PropertyMass"):
    mass: float = 0.0
    inertiaXX: float = 0.0
    inertiaYY: float = 0.0
    inertiaZZ: float = 0.0
    inertiaXY: float = 0.0
    inertiaYZ: float = 0.0
    inertiaZX: float = 0.0
    massModified: Optional[bool] = None
    massModifierX: float = 0.0
    massModifierY: float = 0.0
    massModifierZ: float = 0.0


class PropertySpring(Property, speckle_type=STRUCTURAL_PROPERTY + ".PropertySpring"):
    springType: Optional[PropertyTypeSpring] = None
    springCurveX: float = 0.0
    stiffnessX: float = 0.0
    springCurveY: float = 0.0
    stiffnessY: float = 0.0
    springCurveZ: float = 0.0
    stiffnessZ: float = 0.0
    springCurveXX: float = 0.0
    stiffnessXX: float = 0.0
    springCurveYY: float = 0.0
    stiffnessYY: float = 0.0
    springCurveZZ: float = 0.0
    stiffnessZZ: float = 0.0
    dampingRatio: float = 0.0
    dampingX: float = 0.0
    dampingY: float = 0.0
    dampingZ: float = 0.0
    dampingXX: float = 0.0
    dampingYY: float = 0.0
    dampingZZ: float = 0.0
    matrix: float = 0.0
    postiveLockup: float = 0.0
    frictionCoefficient: float = 0.0


class ReferenceSurfaceEnum(int, Enum):
    Concrete = 0
    Steel = 1
    Timber = 2
    Aluminium = 3
    Masonry = 4
    FRP = 5
    Glass = 6
    Fabric = 7
    Rebar = 8
    Tendon = 9
    ColdFormed = 10
    Other = 11


class shapeType(int, Enum):
    Concrete = 0
    Steel = 1
    Timber = 2
    Aluminium = 3
    Masonry = 4
    FRP = 5
    Glass = 6
    Fabric = 7
    Rebar = 8
    Tendon = 9
    ColdFormed = 10
    Other = 11



from threading import Lock
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from gql.client import Client
from gql.transport.exceptions import TransportQueryError
from graphql import DocumentNode

from specklepy.core.api.credentials import Account
from specklepy.logging.exceptions import (
    GraphQLException,
    SpeckleException,
    UnsupportedException,
)
from specklepy.serialization.base_object_serializer import BaseObjectSerializer
from specklepy.transports.sqlite import SQLiteTransport


class ResourceBase(object):
    def __init__(
        self,
        account: Account,
        basepath: str,
        client: Client,
        name: str,
        server_version: Optional[Tuple[Any, ...]] = None,
    ) -> None:
        self.account = account
        self.basepath = basepath
        self.client = client
        self.name = name
        self.server_version = server_version
        self.schema: Optional[Type] = None
        self.__lock = Lock()

    def _step_into_response(self, response: dict, return_type: Union[str, List, None]):
        """Step into the dict to get the relevant data"""
        if return_type is None:
            return response
        if isinstance(return_type, str):
            return response[return_type]
        if isinstance(return_type, List):
            for key in return_type:
                response = response[key]
            return response

    def _parse_response(self, response: Union[dict, list, None], schema=None):
        """Try to create a class instance from the response"""
        if response is None:
            return None
        if isinstance(response, list):
            return [self._parse_response(response=r, schema=schema) for r in response]
        if schema:
            return schema.model_validate(response)
        elif self.schema:
            try:
                return self.schema.model_validate(response)
            except Exception:
                s = BaseObjectSerializer(read_transport=SQLiteTransport())
                return s.recompose_base(response)
        else:
            return response

    def make_request(
        self,
        query: DocumentNode,
        params: Optional[Dict] = None,
        return_type: Union[str, List, None] = None,
        schema=None,
        parse_response: bool = True,
    ) -> Any:
        """Executes the GraphQL query"""
        try:
            with self.__lock:
                response = self.client.execute(query, variable_values=params)
        except Exception as ex:
            if isinstance(ex, TransportQueryError):
                return GraphQLException(
                    message=(
                        f"Failed to execute the GraphQL {self.name} request. Errors:"
                        f" {ex.errors}"
                    ),
                    errors=ex.errors,
                    data=ex.data,
                )
            else:
                return SpeckleException(
                    message=(
                        f"Failed to execute the GraphQL {self.name} request. Inner"
                        f" exception: {ex}"
                    ),
                    exception=ex,
                )

        response = self._step_into_response(response=response, return_type=return_type)

        if parse_response:
            return self._parse_response(response=response, schema=schema)
        else:
            return response

    def _check_server_version_at_least(
        self, target_version: Tuple[Any, ...], unsupported_message: Optional[str] = None
    ):
        """Use this check to guard against making unsupported requests on older servers.

        Arguments:
            target_version {tuple}
            the minimum server version in the format (major, minor, patch, (tag, build))
            eg (2, 6, 3) for a stable build and (2, 6, 4, 'alpha', 4711) for alpha
        """
        if not unsupported_message:
            unsupported_message = (
                "The client method used is not supported on Speckle Server versions"
                f" prior to v{'.'.join(target_version)}"
            )
        # if version is dev, it should be supported... (or not)
        if self.server_version == ("dev",):
            return
        if self.server_version and self.server_version < target_version:
            raise UnsupportedException(unsupported_message)

    def _check_invites_supported(self):
        """Invites are only supported for Speckle Server >= 2.6.4.
        Use this check to guard against making unsupported requests on older servers.
        """
        self._check_server_version_at_least(
            (2, 6, 4),
            "Stream invites are only supported as of Speckle Server v2.6.4. Please"
            " update your Speckle Server to use this method or use the"
            " `grant_permission` flow instead.",
        )



from functools import wraps
from typing import Callable, Dict, List, Optional, Union

from gql import gql
from graphql import DocumentNode

from specklepy.api.resource import ResourceBase
from specklepy.api.resources.stream import Stream
from specklepy.logging.exceptions import SpeckleException

from specklepy.logging import metrics
from specklepy.core.api.resources.subscriptions import Resource as CoreResource

def check_wsclient(function):
    @wraps(function)
    async def check_wsclient_wrapper(self, *args, **kwargs):
        if self.client is None:
            raise SpeckleException(
                "You must authenticate before you can subscribe to events"
            )
        else:
            return await function(self, *args, **kwargs)

    return check_wsclient_wrapper


class Resource(CoreResource):
    """API Access class for subscriptions"""

    def __init__(self, account, basepath, client) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
        )

    @check_wsclient
    async def stream_added(self, callback: Optional[Callable] = None):
        """Subscribes to new stream added event for your profile.
        Use this to display an up-to-date list of streams.

        Arguments:
            callback {Callable[Stream]} -- a function that takes the updated stream
            as an argument and executes each time a stream is added

        Returns:
            Stream -- the update stream
        """
        metrics.track(metrics.SDK, self.account, {"name": "Subscription Stream Added"})
        return super().stream_added(callback)

    @check_wsclient
    async def stream_updated(self, id: str, callback: Optional[Callable] = None):
        """
        Subscribes to stream updated event.
        Use this in clients/components that pertain only to this stream.

        Arguments:
            id {str} -- the stream id of the stream to subscribe to
            callback {Callable[Stream]}
            -- a function that takes the updated stream
            as an argument and executes each time the stream is updated

        Returns:
            Stream -- the update stream
        """
        metrics.track(metrics.SDK, self.account, {"name": "Subscription Stream Updated"})
        return super().stream_updated(id, callback)

    @check_wsclient
    async def stream_removed(self, callback: Optional[Callable] = None):
        """Subscribes to stream removed event for your profile.
        Use this to display an up-to-date list of streams for your profile.
        NOTE: If someone revokes your permissions on a stream,
        this subscription will be triggered with an extra value of revokedBy
        in the payload.

        Arguments:
            callback {Callable[Dict]}
            -- a function that takes the returned dict as an argument
            and executes each time a stream is removed

        Returns:
            dict -- dict containing 'id' of stream removed and optionally 'revokedBy'
        """
        metrics.track(metrics.SDK, self.account, {"name": "Subscription Stream Removed"})
        return super().stream_removed(callback)

    @check_wsclient
    async def subscribe(
        self,
        query: DocumentNode,
        params: Optional[Dict] = None,
        callback: Optional[Callable] = None,
        return_type: Optional[Union[str, List]] = None,
        schema=None,
        parse_response: bool = True,
    ):
        # if self.client.transport.websocket is None:
        # TODO: add multiple subs to the same ws connection
        async with self.client as session:
            async for res in session.subscribe(query, variable_values=params):
                res = self._step_into_response(response=res, return_type=return_type)
                if parse_response:
                    res = self._parse_response(response=res, schema=schema)
                if callback is not None:
                    callback(res)
                else:
                    return res
